@@ -4,11 +4,13 @@ use base 'CGI::Application';
 sub setup {
     my $self = shift;
     $self->start_mode('rm1');
-    $self->run_modes(
-        rm1 => 'rm1',
-        rm2 => 'rm2',
-        rm3 => 'rm3',
-    ); 
+    $self->run_modes([qw/
+        rm1
+        rm2
+        rm3
+        rm4
+        local_args_to_new
+    /]); 
 }
 
 sub rm1 {
@@ -27,5 +29,16 @@ sub rm3 {
     return "MyApp::Module::Name->rm3 my_param=$param";
 }
 
+# because of caching, we can't re-use PATH_INFO, so we do this. 
+sub rm4 {
+    my $self = shift;
+    return $self->rm3;
+}
+
+sub local_args_to_new {
+    my $self = shift;
+    return $self->tmpl_path;
+}
+
 
 1;
